@@ -65,12 +65,14 @@ static SEXP mkContext(cl_context ctx) {
     return ptr;
 }
 
+#if 0 /* currently unused so disable for now to avoid warnings ... */
 static cl_context getContext(SEXP ctx) {
     if (!Rf_inherits(ctx, "clContext") ||
 	TYPEOF(ctx) != EXTPTRSXP)
 	Rf_error("invalid OpenCL context");
     return (cl_context)R_ExternalPtrAddr(ctx);
 }
+#endif
 
 static void clFreeKernel(SEXP k) {
     clReleaseKernel((cl_kernel)R_ExternalPtrAddr(k));
@@ -230,7 +232,6 @@ SEXP ocl_ez_kernel(SEXP device, SEXP k_name, SEXP code, SEXP prec) {
     cl_device_id device_id = getDeviceID(device);
     cl_program program;
     cl_kernel kernel;
-    int ftype;
 
     if (TYPEOF(k_name) != STRSXP || LENGTH(k_name) != 1)
 	Rf_error("invalid kernel name");
@@ -238,7 +239,6 @@ SEXP ocl_ez_kernel(SEXP device, SEXP k_name, SEXP code, SEXP prec) {
 	Rf_error("invalid kernel code");
     if (TYPEOF(prec) != STRSXP || LENGTH(prec) != 1)
 	Rf_error("invalid precision specification");
-    ftype = (CHAR(STRING_ELT(prec, 0))[0] == 'd') ? FT_DOUBLE : FT_SINGLE;
     ctx = clCreateContext(0, 1, &device_id, NULL, NULL, &err);
     if (!ctx)
 	ocl_err("clCreateContext");
@@ -345,6 +345,12 @@ static SEXP protected_args(struct arg_chain *chain, afin_t fin) {
 typedef struct ocl_call_context {
     cl_mem output;
     cl_command_queue commands;
+    cl_event event;
+    int finished;   /* finished - results have been retrieved */
+#if USE_OCL_COMPLETE_CALLBACK
+    int completed;  /* completed - event has been triggered but not picked up */
+#endif
+    int ftres, ftype, on; /* these are only set is the context leaves ocl_call */
     void *float_out;
     struct arg_chain *float_args, *mem_objects;
 } ocl_call_context_t;
@@ -352,6 +358,9 @@ typedef struct ocl_call_context {
 static void ocl_call_context_fin(SEXP context) {
     ocl_call_context_t *ctx = (ocl_call_context_t*) R_ExternalPtrAddr(context);
     if (ctx) {
+	/* if this was an asynchronous call, we must wait for it to finish */
+	if (!ctx->finished) clFinish(ctx->commands);
+	if (ctx->event) clReleaseEvent(ctx->event);
 	if (ctx->output) clReleaseMemObject(ctx->output);
 	if (ctx->float_args) arg_free(ctx->float_args, 0);
 	if (ctx->float_out) free(ctx->float_out);
@@ -362,10 +371,20 @@ static void ocl_call_context_fin(SEXP context) {
     }
 }
 
+#if USE_OCL_COMPLETE_CALLBACK /* we are curretnly not using the callback since it raises memory management issues and increases complexity */
+static void CL_CALLBACK ocl_complete_callback(cl_event event, cl_int status, void *ucc) {
+    ocl_call_context_t *ctx = (ocl_call_context_t*) ucc;
+    if (ctx) { /* just signal completion - we could be on any thread, so we don't want to issue a callback into R */
+	ctx->completed = 1;
+    }
+}
+#endif
+
+/* .External */
 SEXP ocl_call(SEXP args) {
     struct arg_chain *float_args = 0;
     ocl_call_context_t *occ;
-    int on, an = 0, ftype = FT_DOUBLE, ftsize, ftres;
+    int on, an = 0, ftype = FT_DOUBLE, ftsize, ftres, async;
     size_t global;
     SEXP ker = CADR(args), olen, arg, res, octx;
     cl_kernel kernel = getKernel(ker);
@@ -382,14 +401,16 @@ SEXP ocl_call(SEXP args) {
     if (TYPEOF(res) == STRSXP && LENGTH(res) == 1 && CHAR(STRING_ELT(res, 0))[0] != 'd')
 	ftype = FT_SINGLE;
     ftsize = (ftype == FT_DOUBLE) ? sizeof(double) : sizeof(float);
-    olen = CAR(args);
+    olen = CAR(args);  /* size */
     args = CDR(args);
     on = Rf_asInteger(olen);
     if (on < 0)
 	Rf_error("invalid output length");
-    ftres = (Rf_asInteger(CAR(args)) == 1) ? 1 : 0;
+    ftres = (Rf_asInteger(CAR(args)) == 1) ? 1 : 0;  /* native.result */
     if (ftype != FT_SINGLE) ftres = 0;
-    args = CDR(args);    
+    args = CDR(args);
+    async = (Rf_asInteger(CAR(args)) == 1) ? 0 : 1;  /* wait */
+    args = CDR(args);
     occ = (ocl_call_context_t*) calloc(1, sizeof(ocl_call_context_t));
     if (!occ) Rf_error("unable to allocate ocl_call context");
     octx = PROTECT(R_MakeExternalPtr(occ, R_NilValue, R_NilValue));
@@ -475,9 +496,24 @@ SEXP ocl_call(SEXP args) {
     }
 
     global = on;
-    if (clEnqueueNDRangeKernel(commands, kernel, 1, NULL, &global, NULL, 0, NULL, NULL) != CL_SUCCESS)
+    if (clEnqueueNDRangeKernel(commands, kernel, 1, NULL, &global, NULL, 0, NULL, async ? &occ->event : NULL) != CL_SUCCESS)
 	Rf_error("Error during kernel execution");
+
+    if (async) { /* asynchronous call -> get out and return the context */
+#if USE_OCL_COMPLETE_CALLBACK
+	clSetEventCallback(occ->event, CL_COMPLETE, ocl_complete_callback, occ);
+#endif
+	clFlush(commands); /* the specs don't guarantee execution unless clFlush is called */
+	occ->ftres = ftres;
+	occ->ftype = ftype;
+	occ->on = on;
+	Rf_setAttrib(octx, R_ClassSymbol, mkString("clCallContext"));
+	UNPROTECT(1);
+	return octx;
+    }
+
     clFinish(commands);
+    occ->finished = 1;
 
     /* we can release input memory objects now */
     if (occ->mem_objects) {
@@ -517,3 +553,69 @@ SEXP ocl_call(SEXP args) {
     UNPROTECT(1);
     return res;
 }
+
+SEXP ocl_collect_call(SEXP octx, SEXP wait) {
+    SEXP res = R_NilValue;
+    ocl_call_context_t *occ;
+    int on;
+    cl_int err;
+
+    if (!Rf_inherits(octx, "clCallContext"))
+	Rf_error("Invalid call context");
+    occ = (ocl_call_context_t*) R_ExternalPtrAddr(octx);
+    if (!occ || occ->finished)
+	Rf_error("The call results have already been collected, they cannot be retrieved twice");
+
+    if (Rf_asInteger(wait) == 0 && occ->event) {
+	cl_int status;
+	if ((err = clGetEventInfo(occ->event, CL_EVENT_COMMAND_EXECUTION_STATUS, sizeof(status), &status, NULL)) != CL_SUCCESS)
+	    Rf_error("OpenCL error 0x%x while querying event object for the supplied context", (int) err);
+	
+	if (status < 0)
+	    Rf_error("Asynchronous call failed with error code 0x%x", (int) -status);
+
+	if (status != CL_COMPLETE)
+	    return R_NilValue;
+    }
+
+    clFinish(occ->commands);
+    occ->finished = 1;
+    
+    /* we can release input memory objects now */
+    if (occ->mem_objects) {
+      arg_free(occ->mem_objects, (afin_t) clReleaseMemObject);
+      occ->mem_objects = 0;
+    }
+    if (occ->float_args) {
+      arg_free(occ->float_args, 0);
+      occ->float_args = 0;
+    }
+
+    on = occ->on;
+    res = occ->ftres ? Rf_allocVector(RAWSXP, on * sizeof(float)) : Rf_allocVector(REALSXP, on);
+    if (occ->ftype == FT_SINGLE) {
+	if (occ->ftres) {
+	    if ((err = clEnqueueReadBuffer( occ->commands, occ->output, CL_TRUE, 0, sizeof(float) * on, RAW(res), 0, NULL, NULL )) != CL_SUCCESS)
+		Rf_error("Unable to transfer result vector (%d float elements, oclError %d)", on, err);
+	    PROTECT(res);
+	    Rf_setAttrib(res, R_ClassSymbol, mkString("clFloat"));
+	    UNPROTECT(1);
+	} else {
+	    /* float - need a temporary buffer */
+	    float *fr = (float*) malloc(sizeof(float) * on);
+	    double *r = REAL(res);
+	    int i;
+	    if (!fr)
+		Rf_error("unable to allocate memory for temporary single-precision output buffer");
+	    occ->float_out = fr;
+	    if ((err = clEnqueueReadBuffer( occ->commands, occ->output, CL_TRUE, 0, sizeof(float) * on, fr, 0, NULL, NULL )) != CL_SUCCESS)
+		Rf_error("Unable to transfer result vector (%d float elements, oclError %d)", on, err);
+	    for (i = 0; i < on; i++)
+		r[i] = fr[i];
+	}
+    } else if ((err = clEnqueueReadBuffer( occ->commands, occ->output, CL_TRUE, 0, sizeof(double) * on, REAL(res), 0, NULL, NULL )) != CL_SUCCESS)
+	Rf_error("Unable to transfer result vector (%d double elements, oclError %d)", on, err);
+
+    ocl_call_context_fin(octx);
+    return res;
+}
